local colObjLists = { OBJ_LIST_GENACTOR, OBJ_LIST_PUSHABLE, OBJ_LIST_SURFACE, OBJ_LIST_DESTRUCTIVE }

local _bhvBlacklist = { id_bhvBowser, id_bhvDoor, id_bhvDoorWarp, id_bhvStarDoor, id_bhvUnlockDoorStar, id_bhvToadMessage,
    id_bhvFireSpitter, id_bhvExplosion }
local bhvBlacklist = {}
for i, bhv in ipairs(_bhvBlacklist) do
    bhvBlacklist[bhv] = true
end
_bhvBlacklist = nil

---@param o Object
---@param o2 Object
local function attack_bully(o, o2)
    o2.oBullyLastNetworkPlayerIndex = o.globalPlayerIndex
    o2.oMoveAngleYaw = o.oMoveAngleYaw
    o2.oForwardVel = 30.0

    o2.oInteractStatus = o2.oInteractStatus | ATTACK_FAST_ATTACK | INT_STATUS_WAS_ATTACKED | INT_STATUS_INTERACTED
end

---@param o Object
---@param o2 Object
local function attack_mrblizzard(o, o2)
    if o2.prevObj then
        o2.prevObj.oAction = 2
        o2.prevObj = nil
        o2.oMrBlizzardHeldObj = nil
    end
    o2.oAction = MR_BLIZZARD_ACT_DEATH
end

---@param o Object
---@param o2 Object
local function attack_bullet(o, o2)
    spawn_mist_particles_with_sound(SOUND_GENERAL2_BOBOMB_EXPLOSION)
    o2.oAction = 4
    o2.oTimer = 0
end

---@param o Object
---@param o2 Object
local function attack_chuckya(o, o2)
    o2.oAction = 2
    o2.oVelY = 30
    o2.oMoveAngleYaw = o.oMoveAngleYaw
    o2.oForwardVel = 25
end

bhvMediumAttacks = {
    [id_bhvSmallBully] = attack_bully,
    [id_bhvBigBully] = attack_bully,
    [id_bhvBigBullyWithMinions] = attack_bully,
    [id_bhvSmallChillBully] = attack_bully,
    [id_bhvBigChillBully] = attack_bully,
    [id_bhvBulletBill] = attack_bullet,
    [id_bhvChuckya] = attack_chuckya,
    [id_bhvMrBlizzard] = attack_mrblizzard,
}

---@param o Object
---@param spAttacksList table<BehaviorId,function>
---@param getTarget boolean?
function obj_process_attacks(o, spAttacksList, getTarget)
    -- players
    local m = nearest_mario_state_to_object(o)
    if m and m.playerIndex == 0 and m.marioObj.globalPlayerIndex ~= o.globalPlayerIndex
        and m.action & (ACT_FLAG_INVULNERABLE | ACT_FLAG_INTANGIBLE) == 0 and m.invincTimer == 0
        and obj_check_hitbox_overlap(m.marioObj, o) then
        if spAttacksList[id_bhvMario] then
            spAttacksList[id_bhvMario](o, m)
        else
            take_damage_and_knock_back(m, o)
        end
        if getTarget then return m.marioObj end
    end
    -- other objects
    for i, list in ipairs(colObjLists) do
        local o2 = obj_get_first(list)
        while o2 do
            if o ~= o2 and o2.oInteractStatus & INT_STATUS_INTERACTED == 0 and o2.oIntangibleTimer == 0 and obj_check_hitbox_overlap(o, o2) then
                local bhv = get_id_from_behavior(o2.behavior)
                if not bhvBlacklist[bhv] then
                    if spAttacksList[bhv] then
                        spAttacksList[bhv](o, o2)
                    else
                        o2.oInteractStatus = o2.oInteractStatus | ATTACK_FAST_ATTACK | INT_STATUS_WAS_ATTACKED |
                            INT_STATUS_INTERACTED
                    end
                    if getTarget then return o2 end
                end
            end
            o2 = obj_get_next(o2)
        end
    end
end