local E_MODEL_ARIO_CLOUD_PLATFORM = smlua_model_util_get_id('ario_cloud_plat_geo')
local COL_ARIO_CLOUD_PLATFORM = smlua_collision_util_get('ario_cloud_plat_collision')

local CAP_ACT_THROWN, CAP_ACT_IDLE, CAP_ACT_BOUNCED, CAP_ACT_PLATFORM, CAP_ACT_RETURN = 0, 1, 2, 3, 4

define_custom_obj_fields({ oArioPlatformScale = 'f32', oArioPlatformShake = 'f32', oInhaledObjModel = 'u32', oInhaledObjScale = 'f32', oInhalerIndex = 'u32' })

---@type table<integer, fun(o:Object,m:MarioState)>
local thrownCapActs = {
    [CAP_ACT_THROWN] = function(o, m)
        object_step_without_floor_orient()

        if o.oTimer > 6 then
            if o.oBehParams == 0 then
                o.oAction = CAP_ACT_IDLE
            else
                spawn_mist_particles_with_sound(SOUND_GENERAL_VANISH_SFX)
                cur_obj_scale(0.1)
                o.oFaceAngleYaw = obj_angle_to_object(o, m.marioObj)
                o.oPosY = o.oPosY + 20
                o.oGraphYOffset = 0
                o.oAction = CAP_ACT_PLATFORM
                o.oTimer = 0
                return
            end
        end
        o.oFaceAngleYaw = o.oFaceAngleYaw + 0x1500
    end,
    [CAP_ACT_IDLE] = function(o, m)
        o.oForwardVel = approach_f32_symmetric(o.oForwardVel, 0.0, 10.0)
        object_step_without_floor_orient()

        if (o.oTimer > 32 and m.controller.buttonDown & X_BUTTON == 0) or o.oTimer > 120 then
            o.oAction = CAP_ACT_RETURN
            o.oTimer = 0
        end

        if obj_check_hitbox_overlap(o, m.marioObj) then
            set_mario_action(m, ACT_FREEFALL, 0)
            m.vel.y = 45
            spawn_mist_particles_variable(1, 0, 25.0)
            play_sound(SOUND_OBJ_MRI_SHOOT, o.header.gfx.cameraToObject)
            o.oAction = CAP_ACT_BOUNCED
            o.oTimer = 0
            return
        end
        o.oFaceAngleYaw = o.oFaceAngleYaw + 0x1500
    end,
    [CAP_ACT_BOUNCED] = function(o)
        o.header.gfx.scale.y = (o.oTimer + 1) / 10
        if o.oTimer > 5 then
            o.oAction = CAP_ACT_RETURN
            o.oTimer = 0
        end
    end,
    [CAP_ACT_PLATFORM] = function(o, m)
        if o.oTimer == 0 then
            obj_set_model_extended(o, E_MODEL_ARIO_CLOUD_PLATFORM)
            o.header.gfx.disableAutomaticShadowPos = true
            o.oForwardVel = 0
            o.oVelY = 0
        end

        load_object_collision_model()

        if m.controller.buttonPressed & X_BUTTON ~= 0 or obj_has_model_extended(m.marioObj, E_MODEL_ARIO_CLOUD) == 0 then
            spawn_mist_particles_with_sound(SOUND_OBJ2_MONTY_MOLE_APPEAR)
            obj_set_model_extended(o, CAPMODELS_ARIO_CLOUD.normal)
            cur_obj_scale(1.0)
            o.header.gfx.disableAutomaticShadowPos = false
            o.oPosY = o.oPosY - 20
            o.oGraphYOffset = 50
            o.oAction = CAP_ACT_RETURN
            o.oTimer = 0
            return
        end

        o.oArioPlatformScale = approach_f32_asymptotic(o.oArioPlatformScale, 1.0, 1.25)
        o.oArioPlatformShake = approach_f32_symmetric(o.oArioPlatformShake, 0.05, 0.01)

        local t = o.oTimer * 1092
        local xz = maxf(0.0, o.oArioPlatformScale - sins(t) * o.oArioPlatformShake)
        local y = maxf(0.0, o.oArioPlatformScale - coss(t) * o.oArioPlatformShake)
        local nearestM = nearest_mario_state_to_object(o)
        if nearestM.marioObj.platform == o then
            --squish
            y = y * 0.85
            if nearestM.action == ACT_ARIO_AIR_ROLL or nearestM.action == ACT_GROUND_POUND_LAND then
                o.oArioPlatformShake = 0.5
                o.oTimer = 1
            end
        end
        obj_scale_xyz(o, xz, y, xz)

        o.header.gfx.shadowPos.x = o.oPosX
        o.header.gfx.shadowPos.y = o.oPosY - 50
        o.header.gfx.shadowPos.z = o.oPosZ
    end,
    [CAP_ACT_RETURN] = function(o, m)
        o.oMoveAngleYaw = obj_angle_to_object(o, m.marioObj)
        o.oForwardVel = approach_f32_symmetric(o.oForwardVel, 120.0, 30.0)

        cur_obj_move_xz_using_fvel_and_yaw()

        o.oPosY = approach_f32_asymptotic(o.oPosY, m.pos.y, 0.25)

        if o.oTimer > 1 and obj_check_hitbox_overlap(o, m.marioObj) then
            play_sound(SOUND_ACTION_HANGING_STEP, o.header.gfx.cameraToObject)
            gPlayerSyncTable[m.playerIndex].capThrown = false
            -- keep players from using the platform throw again until landing
            if m.action & ACT_FLAG_AIR ~= 0 then gStateExtras[m.playerIndex].restrictCapThrow = true end
            obj_mark_for_deletion(o)
        end
        o.oFaceAngleYaw = o.oFaceAngleYaw + 0x1500
    end
}

---@param o Object
local function thrown_cap_init(o)
    o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
    o.oGraphYOffset = 50
    cur_obj_scale(1.0)
    o.oOpacity = 0xFF

    cur_obj_set_hitbox_radius_and_height(75, 50)
    o.hitboxDownOffset = -50
    cur_obj_set_hurtbox_radius_and_height(75, 50)

    o.oGravity = 0.0
    o.oDragStrength = 0.0
    o.oBuoyancy = 1.0
    o.oFriction = 1.0
    o.oBounciness = 0.0

    o.collisionData = COL_ARIO_CLOUD_PLATFORM
    o.oCollisionDistance = 1000

    network_init_object(o, true, { 'oArioPlatformScale', 'oArioPlatformShake' })

    o.oArioPlatformScale = 0.01
    o.oArioPlatformShake = 0.05
end
---@param o Object
local function thrown_cap_loop(o)
    local m = gMarioStates[network_local_index_from_global(o.globalPlayerIndex)]

    thrownCapActs[o.oAction](o, m)

    if o.header.gfx.areaIndex ~= m.area.index or not gPlayerSyncTable[m.playerIndex].capThrown then
        o.header.gfx.disableAutomaticShadowPos = false
        gPlayerSyncTable[m.playerIndex].capThrown = false
        obj_mark_for_deletion(o)
    end
end

---@param o Object
local function inhaled_obj_init(o)
    o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE | OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW | OBJ_FLAG_HOLDABLE

    o.hitboxRadius = 40
    o.hitboxHeight = 60
    o.hitboxDownOffset = 0

    o.oGravity = 0.0
    o.oDragStrength = 0.0
    o.oBuoyancy = 1.0
    o.oFriction = 1.0
    o.oBounciness = 0.0
    o.oInteractType = INTERACT_GRABBABLE

    cur_obj_become_intangible()

    cur_obj_init_animation(0)

    network_init_object(o, true, { 'oInhaledObjModel', 'oInhaledObjScale', 'oInhalerIndex' })
end

---@param o Object
local function inhaled_obj_loop(o)
    local m = gMarioStates[network_local_index_from_global(o.oInhalerIndex)]

    if o.oHeldState == HELD_FREE then
        if o.oBehParams == 0 then
            -- being sucked

            o.oMoveAngleYaw = obj_angle_to_point(o, m.pos.x, m.pos.z)
            o.oForwardVel = approach_f32_symmetric(o.oForwardVel, 30.0, 3.0)
            o.oPosY = approach_f32_asymptotic(o.oPosY, m.pos.y + 40, 1.5)

            cur_obj_move_using_fvel_and_gravity()

            cur_obj_scale(maxf(0.0, o.header.gfx.scale.x - 0.025))

            if dist_between_objects(o, m.marioObj) < 100 or o.oTimer > 50 then
                o.oForwardVel = 0.0
                m.interactObj = o
                m.usedObj = o
                mario_grab_used_object(m)

                if m.action & ACT_GROUP_MASK == ACT_GROUP_AIRBORNE then
                    set_mario_action(m, ACT_HOLD_FREEFALL, 0)
                else
                    set_mario_action(m, ACT_HOLD_IDLE, 0)
                end

                obj_set_model_extended(o, E_MODEL_NONE)

                spawn_mist_particles_variable(3, 0, 16.0)
                play_sound(SOUND_ACTION_UNSTUCK_FROM_GROUND, o.header.gfx.cameraToObject)
            end
        else
            -- being shot
            o.oForwardVel = 40.0

            cur_obj_update_floor_and_resolve_wall_collisions(45)
            cur_obj_move_standard(45)

            cur_obj_scale(minf(o.header.gfx.scale.x + 0.4, o.oInhaledObjScale))

            obj_process_attacks(o, bhvMediumAttacks)

            if o.oTimer > 30 or o.oMoveFlags & OBJ_MOVE_HIT_WALL ~= 0 then
                spawn_mist_particles()
                obj_mark_for_deletion(o)
                play_sound(SOUND_OBJ_ENEMY_DEATH_HIGH, o.header.gfx.cameraToObject)
            end
        end
    elseif o.oHeldState == HELD_HELD then
        cur_obj_disable_rendering()
        cur_obj_become_intangible()
        cur_obj_set_pos_relative(m.marioObj, 30, 80, 100)
        o.oMoveAngleYaw = m.faceAngle.y
        cur_obj_scale(0.0)

        o.oTimer = 0
        o.oBehParams = 1
    elseif o.oHeldState == HELD_THROWN then
        cur_obj_enable_rendering()
        cur_obj_become_tangible()
        cur_obj_set_pos_relative(m.marioObj, 30, 80, 100)
        o.oMoveAngleYaw = m.faceAngle.y

        obj_set_model_extended(o, o.oInhaledObjModel)

        o.oTimer = 0
        o.oBehParams = 1
        o.oHeldState = HELD_FREE
        o.oInteractType = 0

        play_sound(SOUND_OBJ_MRI_SHOOT, o.header.gfx.cameraToObject)
    else
        cur_obj_set_pos_relative(m.marioObj, 30, 60, 100)
        o.oHeldState = HELD_FREE
        spawn_mist_particles()
        obj_mark_for_deletion(o)
    end

    o.oFaceAnglePitch = o.oFaceAnglePitch + 0xF00
end

---@param o Object
local function inhaled_player_init(o)
    o.oFlags = OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW | OBJ_FLAG_HOLDABLE

    o.hitboxRadius = 30
    o.hitboxHeight = 60
    o.hitboxDownOffset = 0

    o.oGravity = 0.0
    o.oDragStrength = 0.0
    o.oBuoyancy = 1.0
    o.oFriction = 1.0
    o.oBounciness = 0.0
    o.oInteractType = INTERACT_GRABBABLE

    cur_obj_become_intangible()

    network_init_object(o, true, {'oInhalerIndex'})
end

---@param o Object
local function inhaled_player_loop(o)
    local m = gMarioStates[network_local_index_from_global(o.oInhalerIndex)]

    if o.oHeldState == HELD_FREE then
        m.interactObj = o
        m.usedObj = o
        mario_grab_used_object(m)

        if m.action & ACT_GROUP_MASK == ACT_GROUP_AIRBORNE then
            set_mario_action(m, ACT_HOLD_FREEFALL, 0)
        else
            set_mario_action(m, ACT_HOLD_IDLE, 0)
        end
    elseif o.oHeldState == HELD_HELD then
        cur_obj_disable_rendering()
        cur_obj_become_intangible()
        cur_obj_set_pos_relative(m.marioObj, 30, 80, 100)
    elseif o.oHeldState == HELD_THROWN then
        cur_obj_enable_rendering()
        cur_obj_become_tangible()
        cur_obj_set_pos_relative(m.marioObj, 30, 80, 100)

        if m.playerIndex == 0 then
            network_send_to(network_local_index_from_global(o.globalPlayerIndex), true,
                { id = PACKET_SHOT_FROM_INHALE, i = network_global_index_from_local(0) })
        end

        obj_mark_for_deletion(o)

        play_sound(SOUND_OBJ_MRI_SHOOT, o.header.gfx.cameraToObject)
    else
        cur_obj_set_pos_relative(m.marioObj, 30, 60, 100)

        if m.playerIndex == 0 then
            network_send_to(network_local_index_from_global(o.globalPlayerIndex), true,
                { id = PACKET_DROPPED_FROM_INHALE, i = network_global_index_from_local(0) })
        end

        obj_mark_for_deletion(o)
    end
end


---@param o Object
local function held_cap_init(o)
    o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE

    o.oFaceAnglePitch = -0x4000

    cur_obj_become_intangible()

    network_init_object(o, true, {})
end

---@param o Object
local function held_cap_loop(o)
    local m = gMarioStates[network_local_index_from_global(o.globalPlayerIndex)]

    o.oPosX = m.pos.x + sins(m.faceAngle.y) * 70
    o.oPosZ = m.pos.z + coss(m.faceAngle.y) * 70
    o.oPosY = m.marioBodyState.handFootPos[1].y or m.pos.y + 80
    o.oFaceAngleYaw = m.faceAngle.y
    if holding_inhaled_obj(m) then
        obj_scale(o, 1.25)
    elseif m.playerIndex == 0 and m.action ~= ACT_ARIO_CAP_INHALE and m.action ~= ACT_ARIO_CAP_INHALE_AIR then
        obj_mark_for_deletion(o)
    end

    cur_obj_unhide()
end


---@param o Object
local function bomb_bit_init(o)
    o.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE | OBJ_FLAG_SET_FACE_YAW_TO_MOVE_YAW | OBJ_FLAG_HOLDABLE

    o.oInteractType = INTERACT_GRABBABLE
    o.hitboxDownOffset = 0
    o.oDamageOrCoinValue = 0
    o.oHealth = 0
    o.oNumLootCoins = 0
    o.hitboxRadius = 65
    o.hitboxHeight = 113

    o.oGravity = 2.5
    o.oFriction = 0.8
    o.oBuoyancy = 1.3
    o.oInteractionSubtype = INT_SUBTYPE_KICKABLE

    o.oBobombFuseLit = 1
    smlua_anim_util_set_animation(o, 'bomb_bit_idle')
    cur_obj_scale(1.0)
    cur_obj_become_tangible()

    network_init_object(o, true, { 'oBobombFuseLit', 'oBobombFuseTimer' })
end

---@type table<integer,fun(o:Object,player?:Object)>
local bombBitActions = {
    [BOBOMB_ACT_CHASE_MARIO] = function(o, player)
        local animInfo = o.header.gfx.animInfo
        animInfo.animFrame = animInfo.animFrame + 1
        o.oForwardVel = 12.0

        local collisionFlags = object_step()

        if animInfo.animFrame == 5 or animInfo.animFrame == 16 then
            cur_obj_play_sound_2(SOUND_OBJ_BOBOMB_WALK)
        end

        if player then
            obj_turn_toward_object(o, player, 16, 0x800)
        end
        if o.oFloor then
            obj_check_floor_death(collisionFlags, o.oFloor)
        end
    end,
    [BOBOMB_ACT_LAUNCHED] = function(o, player)
        local collisionFlags = object_step()
        if (collisionFlags & OBJ_COL_FLAG_GROUNDED) == OBJ_COL_FLAG_GROUNDED then
            o.oAction = BOBOMB_ACT_EXPLODE
            o.oTimer = 0
        end
    end,
    [BOBOMB_ACT_EXPLODE] = function(o, player)
        if o.oTimer < 4 then
            cur_obj_scale(1.0 + o.oTimer / 4.0)
        else
            local explosion = spawn_sync_object(id_bhvSparkle, E_MODEL_EXPLOSION, o.oPosX, o.oPosY, o.oPosZ, nil)
            play_sound(SOUND_OBJ_BULLY_EXPLODE, o.header.gfx.cameraToObject)
            if explosion then
                explosion.oGraphYOffset = 100.0
                obj_scale(explosion, 0.5)
                explosion.oDamageOrCoinValue = 0
            end

            o.activeFlags = ACTIVE_FLAG_DEACTIVATED
        end
    end,
    [OBJ_ACT_LAVA_DEATH] = function(o, player)
        obj_lava_death()
    end,
    [OBJ_ACT_DEATH_PLANE_DEATH] = function(o, player)
        o.activeFlags = ACTIVE_FLAG_DEACTIVATED
    end
}

---@param o Object
local function bomb_bit_loop(o)
    if is_point_within_radius_of_mario(o.oPosX, o.oPosY, o.oPosZ, 4000) ~= 0 then
        if o.oHeldState == HELD_FREE then
            local player = nearest_player_to_object(o)

            local act = bombBitActions[o.oAction]
            if act then act(o, player) end

            if o.oInteractStatus & INT_STATUS_INTERACTED ~= 0 then
                if o.oInteractStatus & INT_STATUS_MARIO_UNK1 ~= 0 then
                    if player then
                        o.oMoveAngleYaw = player.header.gfx.angle.y
                    end
                    o.oForwardVel = 25.0
                    o.oVelY = 30.0
                    o.oAction = BOBOMB_ACT_LAUNCHED
                end
                if o.oInteractStatus & INT_STATUS_TOUCHED_BOB_OMB ~= 0 then
                    o.oAction = BOBOMB_ACT_EXPLODE
                end
                o.oInteractStatus = 0
            end
        elseif o.oHeldState == HELD_HELD then
            cur_obj_hide()

            local player = gMarioStates[o.heldByPlayerIndex].marioObj
            cur_obj_set_pos_relative(player, 0, 60.0, 100.0)
        elseif o.oHeldState == HELD_THROWN then
            cur_obj_enable_rendering_2()

            cur_obj_unhide()
            o.oHeldState = HELD_FREE
            o.oFlags = o.oFlags & ~0x8
            o.oForwardVel = 25.0
            o.oVelY = 20.0
            o.oAction = BOBOMB_ACT_LAUNCHED
        else
            cur_obj_get_dropped()

            cur_obj_unhide()

            o.oHeldState = HELD_FREE
            o.oAction = BOBOMB_ACT_CHASE_MARIO
        end

        local m = gMarioStates[network_local_index_from_global(o.globalPlayerIndex)]
        if m.playerIndex ~= 0 then return end
        if m.controller.buttonPressed & L_TRIG ~= 0 and m.controller.buttonDown & Z_TRIG ~= 0 then
            o.oAction = BOBOMB_ACT_EXPLODE
            o.oTimer = 0
            network_send_object(o, true)
        end
        if o.activeFlags == ACTIVE_FLAG_DEACTIVATED then
            gStateExtras[network_local_index_from_global(o.globalPlayerIndex)].bombBitSpawned = false
        end
    end
end


id_bhvArioThrownCap = hook_behavior(nil, OBJ_LIST_SURFACE, false, thrown_cap_init, thrown_cap_loop, 'bhvArioThrownCap')
id_bhvCapInhaledObj = hook_behavior(nil, OBJ_LIST_GENACTOR, false, inhaled_obj_init, inhaled_obj_loop, 'bhvCapInhaledObj')
id_bhvCapInhaledPlayer = hook_behavior(nil, OBJ_LIST_GENACTOR, false, inhaled_player_init, inhaled_player_loop,
    'bhvCapInhaledPlayer')
id_bhvArioHeldCap = hook_behavior(nil, OBJ_LIST_UNIMPORTANT, false, held_cap_init, held_cap_loop, 'bhvArioHeldCap')
id_bhvBombBit = hook_behavior(nil, OBJ_LIST_GENACTOR, false, bomb_bit_init, bomb_bit_loop, 'bhvBombBit')
