ACT_ARIO_SPIN = allocate_mario_action(ACT_GROUP_MOVING | ACT_FLAG_MOVING | ACT_FLAG_ATTACKING)
ACT_ARIO_AIR_SPIN = allocate_mario_action(ACT_GROUP_AIRBORNE | ACT_FLAG_AIR | ACT_FLAG_ATTACKING |
    ACT_FLAG_ALLOW_VERTICAL_WIND_ACTION)
ACT_ARIO_WATER_SPIN = allocate_mario_action(ACT_GROUP_SUBMERGED | ACT_FLAG_MOVING | ACT_FLAG_SWIMMING |
    ACT_FLAG_SWIMMING_OR_FLYING | ACT_FLAG_WATER_OR_TEXT | ACT_FLAG_ATTACKING)
ACT_ARIO_CAP_THROW = allocate_mario_action(ACT_GROUP_AIRBORNE | ACT_FLAG_MOVING | ACT_FLAG_AIR)
ACT_ARIO_ROLL = allocate_mario_action(ACT_GROUP_MOVING | ACT_FLAG_MOVING | ACT_FLAG_ATTACKING)
ACT_ARIO_WATER_ROLL = allocate_mario_action(ACT_GROUP_AIRBORNE | ACT_FLAG_MOVING | ACT_FLAG_ATTACKING)
ACT_ARIO_AIR_ROLL = allocate_mario_action(ACT_GROUP_AIRBORNE | ACT_FLAG_AIR | ACT_FLAG_ATTACKING |
    ACT_FLAG_ALLOW_VERTICAL_WIND_ACTION)
ACT_ARIO_CAP_HIDE = allocate_mario_action(ACT_GROUP_OBJECT | ACT_FLAG_MOVING | ACT_FLAG_SHORT_HITBOX)
ACT_ARIO_CAP_HIDE_WATER = allocate_mario_action(ACT_GROUP_SUBMERGED | ACT_FLAG_SHORT_HITBOX)
ACT_ARIO_CAP_INHALE = allocate_mario_action(ACT_GROUP_MOVING | ACT_FLAG_MOVING)
ACT_ARIO_CAP_INHALE_AIR = allocate_mario_action(ACT_GROUP_AIRBORNE | ACT_FLAG_AIR)
ACT_INHALED = allocate_mario_action(ACT_GROUP_AUTOMATIC | ACT_FLAG_INTANGIBLE)

local SFX_STAR = audio_sample_load('sfx_star.ogg')

---@param m MarioState
local function ario_spin_loop(m)
    if m.actionTimer == 0 then
        m.forwardVel = math.max(math.min(m.forwardVel + 10.0, 50.0), m.forwardVel)
        set_mario_animation(m, CHAR_ANIM_TWIRL)
        audio_sample_play(SFX_STAR, m.pos, 1.0)
    end
    smlua_anim_util_set_animation(m.marioObj, 'ario_spin_ground')
    m.marioBodyState.eyeState = MARIO_EYES_LOOK_UP

    local angleDiff = abs_angle_diff(m.intendedYaw, m.faceAngle.y)
    if m.input & INPUT_NONZERO_ANALOG ~= 0 then
        m.forwardVel = math.max(5.0, m.forwardVel - 1.0 * (angleDiff / 0x8000))
    else
        m.forwardVel = math.max(5.0, m.forwardVel - 1.0)
    end

    if angleDiff < 0x6000 then
        m.faceAngle.y = approach_s16_symmetric(m.faceAngle.y, m.intendedYaw, 0x280 + m.actionTimer * 16)
    end

    mario_set_forward_vel(m, m.forwardVel)
    if abs_angle_diff(m.floorAngle, m.faceAngle.y) > 0x4000 then
        local surface = mario_get_floor_class(m)
        if surface == SURFACE_CLASS_VERY_SLIPPERY then
            m.forwardVel = approach_f32_symmetric(m.forwardVel, 0.0, 3.0)
        elseif surface == SURFACE_CLASS_SLIPPERY then
            m.forwardVel = approach_f32_symmetric(m.forwardVel, 0.0, 1.0)
        end
    end

    local step = perform_ground_step(m)

    if m.actionTimer >= 30 then
        return set_mario_action(m, ACT_WALKING, 0)
    end
    if m.actionTimer > 3 then
        if m.controller.buttonPressed & A_BUTTON ~= 0 then
            return set_mario_action(m, ACT_DOUBLE_JUMP, 0)
        end
        if m.controller.buttonPressed & X_BUTTON ~= 0 and not gPlayerSyncTable[m.playerIndex].capThrown and m.flags & MARIO_NORMAL_CAP ~= 0 then
            return set_mario_action(m, ACT_ARIO_CAP_THROW, 0)
        end
    end
    if step == GROUND_STEP_HIT_WALL then
        return slide_bonk(m, ACT_BACKWARD_GROUND_KB, ACT_BRAKING_STOP)
    elseif step == GROUND_STEP_LEFT_GROUND then
        return set_mario_action(m, ACT_FREEFALL, 0)
    end

    m.marioObj.header.gfx.angle.x = 0x0500

    if m.actionTimer <= 20 then
        m.marioBodyState.handState = 1
        set_mario_particle_flags(m, PARTICLE_SPARKLES, 0)
    end

    if m.actionTimer == 1 or m.actionTimer == 6 then
        play_sound(SOUND_GENERAL_SWISH_WATER, m.marioObj.header.gfx.cameraToObject)
    end

    m.actionTimer = m.actionTimer + 1
end

---@param m MarioState
local function ario_air_spin_loop(m)
    if m.actionTimer == 0 then
        m.vel.y = 25.0 * (1.0 + m.numStars / 120)
        set_mario_animation(m, CHAR_ANIM_TWIRL)
        audio_sample_play(SFX_STAR, m.pos, 1.0)
    end
    smlua_anim_util_set_animation(m.marioObj, 'ario_spin_air')
    m.marioBodyState.eyeState = MARIO_EYES_LOOK_UP

    local step = common_air_action_step(m, ACT_FREEFALL_LAND, CHAR_ANIM_TWIRL, AIR_STEP_CHECK_LEDGE_GRAB)

    if m.actionTimer >= 20 then
        return set_mario_action(m, ACT_FREEFALL, 0)
    end
    if m.controller.buttonPressed & X_BUTTON ~= 0 and m.playerIndex == 0 and m.actionTimer > 3 and not gPlayerSyncTable[0].capThrown and m.flags & MARIO_NORMAL_CAP ~= 0 then
        if _G.charSelect.character_get_current_costume(0) == 1 then
            if not gStateExtras[0].restrictCapThrow then
                gPlayerSyncTable[0].capThrown = true
                spawn_sync_object(id_bhvArioThrownCap, CAPMODELS_ARIO_CLOUD.normal, m.pos.x, m.pos.y + 20, m.pos.z,
                    ---@param o Object
                    function(o)
                        o.globalPlayerIndex = m.marioObj.globalPlayerIndex
                        o.oMoveAngleYaw = m.faceAngle.y
                        o.oVelY = -40.0
                        o.oAction = 0
                        -- skip 3 frames of the timer so the platform appears quicker
                        o.oTimer = 3
                        o.oBehParams = 1
                        play_sound(SOUND_ACTION_THROW, o.header.gfx.cameraToObject)
                    end)
                m.vel.y = 8.0
                m.forwardVel = 0.0
            end
        else
            return set_mario_action(m, ACT_ARIO_CAP_THROW, 1)
        end
    end

    if m.actionTimer <= 12 then
        m.marioBodyState.handState = 1
        set_mario_particle_flags(m, PARTICLE_SPARKLES, 0)
    end

    if m.actionTimer == 1 or m.actionTimer == 5 then
        play_sound(SOUND_GENERAL_SWISH_WATER, m.marioObj.header.gfx.cameraToObject)
    end

    m.actionTimer = m.actionTimer + 1
end

---@param m MarioState
local function ario_air_spin_gravity(m)
    m.vel.y = m.vel.y - 3.0
end

---@param m MarioState
local function ario_water_spin_loop(m)
    if m.actionTimer == 0 then
        m.forwardVel = 30.0 * (1.0 + m.numStars / 240)
        set_mario_animation(m, CHAR_ANIM_TWIRL)
        audio_sample_play(SFX_STAR, m.pos, 1.0)
    end
    smlua_anim_util_set_animation(m.marioObj, 'ario_spin_water')

    update_flying(m)
    perform_water_step(m)

    if m.actionTimer >= 30 then
        return set_mario_action(m, ACT_WATER_IDLE, 0)
    end

    if m.actionTimer <= 16 then
        m.marioBodyState.handState = 1
        set_mario_particle_flags(m, PARTICLE_SPARKLES, 0)

        if m.pos.y >= m.waterLevel - 80 and m.faceAngle.x > 0x1800 then
            m.pos.y = m.waterLevel
            local vel = m.forwardVel * 1.5
            m.vel.y = sins(m.faceAngle.x) * vel
            m.forwardVel = coss(m.faceAngle.x) * vel
            return set_mario_action(m, ACT_DIVE, 0)
        end
    end

    if m.actionTimer == 1 or m.actionTimer == 5 then
        play_sound(SOUND_GENERAL_SWISH_WATER, m.marioObj.header.gfx.cameraToObject)
    end

    m.actionTimer = m.actionTimer + 1
end

hook_mario_action(ACT_ARIO_SPIN, ario_spin_loop, INT_SLIDE_KICK)
hook_mario_action(ACT_ARIO_AIR_SPIN, { every_frame = ario_air_spin_loop, gravity = ario_air_spin_gravity },
    INT_SLIDE_KICK)
hook_mario_action(ACT_ARIO_WATER_SPIN, ario_water_spin_loop)


---@param m MarioState
local function ario_cap_throw_loop(m)
    set_mario_animation(m, CHAR_ANIM_THROW_LIGHT_OBJECT)
    if m.actionTimer == 2 then
        if m.playerIndex == 0 then
            gPlayerSyncTable[0].capThrown = true
            local alt = _G.charSelect.character_get_current_costume(0)
            ---@param o Object
            spawn_sync_object(id_bhvArioThrownCap, alt == 1 and CAPMODELS_ARIO_CLOUD.normal or CAPMODELS_ARIO.normal, m.pos.x, m.pos.y + 20, m.pos.z, function(o)
                o.globalPlayerIndex = m.marioObj.globalPlayerIndex
                o.oMoveAngleYaw = m.faceAngle.y
                o.oForwardVel = 50.0 + m.forwardVel * 0.8
                o.oAction = 0
                if alt == 1 then
                    o.oBehParams = 1
                end
                play_sound(SOUND_ACTION_THROW, o.header.gfx.cameraToObject)
            end)
        end
    end

    if m.actionArg == 0 then
        -- grounded

        m.forwardVel = approach_f32_symmetric(m.forwardVel, 0.0, 2.0)
        local step = perform_ground_step(m)

        if m.actionTimer > 5 then
            if m.input & INPUT_NONZERO_ANALOG ~= 0 then
                return set_mario_action(m, ACT_WALKING, 0)
            end
            if m.actionTimer > 10 then
                return set_mario_action(m, ACT_WALKING, 0)
            end
        end

        if step == GROUND_STEP_LEFT_GROUND then
            return set_mario_action(m, ACT_FREEFALL, 0)
        end
    else
        --aerial

        update_air_without_turn(m)
        local step = perform_air_step(m, 0)

        if m.actionTimer > 3 then
            if m.controller.buttonPressed & Z_TRIG ~= 0 then
                return set_mario_action(m, ACT_GROUND_POUND, 0)
            elseif m.controller.buttonPressed & B_BUTTON ~= 0 then
                return set_mario_action(m, ACT_DIVE, 0)
            elseif m.controller.buttonPressed & Y_BUTTON ~= 0 and not gStateExtras[m.playerIndex].airSpinUsed then
                return set_mario_action(m, ACT_ARIO_AIR_SPIN, 0)
            end
            if m.actionTimer > 6 then
                return set_mario_action(m, ACT_FREEFALL, 0)
            end
        end
        if step == AIR_STEP_LANDED then
            return set_mario_action(m, ACT_FREEFALL_LAND, 0)
        end
    end
    m.actionTimer = m.actionTimer + 1
end
---@param m MarioState
local function ario_cap_throw_gravity(m)
    if m.actionArg == 1 then
        m.vel.y = m.vel.y - 2.0
    end
end

hook_mario_action(ACT_ARIO_CAP_THROW, { every_frame = ario_cap_throw_loop, gravity = ario_cap_throw_gravity })

---@param m MarioState
local function ario_roll_loop(m)
    local intendedDYaw = m.intendedYaw - m.slideYaw
    local forward = coss(intendedDYaw)
    local sideward = sins(intendedDYaw)

    if m.input & INPUT_NONZERO_ANALOG == 0 then
        forward = 0.5
    end

    local turnFactor = sideward * (1 / maxf(1, m.forwardVel) * 3)
    local oldX = m.slideVelX
    m.slideVelX = m.slideVelX + m.slideVelZ * (m.intendedMag / 32.0) * turnFactor;
    m.slideVelZ = m.slideVelZ - oldX * (m.intendedMag / 32.0) * turnFactor;

    if m.forwardVel < 10.0 then
        m.slideVelX = m.slideVelX + (m.intendedMag / 32.0) * sins(m.intendedYaw)
        m.slideVelZ = m.slideVelZ + (m.intendedMag / 32.0) * coss(m.intendedYaw)
    end
    if m.forwardVel > 0.02 then
        update_sliding_angle(m, 7.0, m.intendedMag / 32 * forward * 0.02 + 0.98)
        m.faceAngle.y = approach_s16_asymptotic(m.faceAngle.y, m.slideYaw, 2)
    else
        m.forwardVel = math.sqrt(sqr(m.slideVelX) + sqr(m.slideVelZ))
    end

    local vel = math.sqrt(sqr(m.slideVelX) + sqr(m.slideVelZ))

    if vel < 8.0 then
        -- slow speed
        if m.controller.buttonDown & Z_TRIG == 0 then
            return set_mario_action(m, ACT_STOP_CROUCHING, 0)
        end
        if mario_floor_is_slippery(m) == 0 then
            if m.controller.buttonPressed & A_BUTTON ~= 0 then
                return set_mario_action(m, ACT_BACKFLIP, 0)
            end
            if m.controller.buttonPressed & B_BUTTON ~= 0 then
                return set_mario_action(m, ACT_PUNCHING, 9)
            end
        end
    else
        -- moving speed
        if m.controller.buttonDown & Z_TRIG == 0 then
            return set_mario_action(m, ACT_BRAKING, 0)
        end
        if mario_floor_is_slippery(m) == 0 or (abs_angle_diff(m.slideYaw, m.floorAngle) < 0x4000) then
            if m.controller.buttonPressed & A_BUTTON ~= 0 then
                return set_mario_action(m, ACT_LONG_JUMP, 0)
            end
            if m.controller.buttonPressed & B_BUTTON ~= 0 then
                return set_mario_action(m, ACT_SLIDE_KICK, 9)
            end
        end

        if vel >= 15.0 and m.pos.y <= m.waterLevel - 25.0 then
            m.pos.y = m.waterLevel
            return set_mario_action(m, ACT_ARIO_WATER_ROLL, 0)
        end

        set_mario_particle_flags(m, PARTICLE_DUST, 0)
        play_sound_with_freq_scale(SOUND_MOVING_TERRAIN_SLIDE + m.terrainSoundAddend,
            m.marioObj.header.gfx.cameraToObject, 1.0 + vel / 64.0)
    end

    if m.actionState <= 0 then
        if m.controller.buttonPressed & Y_BUTTON ~= 0 then
            play_sound_with_freq_scale(SOUND_OBJ_MRI_SHOOT, m.marioObj.header.gfx.cameraToObject, 1.25)
            audio_sample_play(SFX_STAR, m.pos, 0.5)
            m.actionState = 22
        end
    elseif m.actionState > 10 then
        local newVel = maxf(50.0, vel)
        m.slideVelX = sins(m.slideYaw) * newVel
        m.slideVelZ = coss(m.slideYaw) * newVel
        set_mario_particle_flags(m, PARTICLE_SPARKLES, 0)
    end
    if m.actionState > 0 then
        m.actionState = m.actionState - 1
    end

    local step = perform_ground_step(m)

    if step == GROUND_STEP_LEFT_GROUND then
        m.vel.y = 8.0
        return set_mario_action(m, ACT_ARIO_AIR_ROLL, 0)
    elseif step == GROUND_STEP_HIT_WALL and m.forwardVel > 5.0 then
        mario_bonk_reflection(m, 0)
        m.slideYaw = m.faceAngle.y
        m.slideVelX = sins(m.slideYaw) * vel
        m.slideVelZ = coss(m.slideYaw) * vel
    end

    if m.actionTimer == 0 then
        if not gPlayerSyncTable[m.playerIndex].capThrown and m.flags & MARIO_NORMAL_CAP ~= 0 then
            if ario_enter_model_swap(m, 2) then
                spawn_mist_particles_variable(4, 1, 32.0)
                play_sound(SOUND_OBJ_BOO_BOUNCE_TOP, m.marioObj.header.gfx.cameraToObject)
            end
        end
    else
        local accel = max(1, math.sqrt(vel) * 0x9000)
        set_character_anim_with_accel(m, CHAR_ANIM_FORWARD_SPINNING, accel)
        if is_ario_ball_form(m) then
            smlua_anim_util_set_animation(m.marioObj, 'ario_ball_roll')
        else
            smlua_anim_util_set_animation(m.marioObj, 'ario_roll')
        end
    end
    align_with_floor(m)

    m.actionTimer = m.actionTimer + 1
end

---@param m MarioState
local function ario_water_roll_loop(m)
    update_air_with_turn(m)

    m.pos.y = m.waterLevel - maxf(50.0 - m.forwardVel, 0.0)

    if m.forwardVel < 15.0 or m.controller.buttonDown & Z_TRIG == 0 then
        return set_mario_action(m, ACT_FREEFALL, 0)
    end

    if m.controller.buttonPressed & A_BUTTON ~= 0 then
        return set_mario_action(m, ACT_LONG_JUMP, 0)
    end
    if m.controller.buttonPressed & B_BUTTON ~= 0 then
        return set_mario_action(m, ACT_SLIDE_KICK, 9)
    end

    local step = perform_air_step(m, 0)

    if step == AIR_STEP_HIT_WALL and gServerSettings.bouncyLevelBounds == 0 then
        mario_bonk_reflection(m, 0)
    elseif step == AIR_STEP_LANDED then
        return set_mario_action(m, ACT_ARIO_ROLL, 0)
    end

    if m.actionTimer == 0 then
        if not gPlayerSyncTable[m.playerIndex].capThrown and m.flags & MARIO_NORMAL_CAP ~= 0 then
            if ario_enter_model_swap(m, 2) then
                play_sound(SOUND_OBJ_BOO_BOUNCE_TOP, m.marioObj.header.gfx.cameraToObject)
            end
        end
    else
        local accel = max(0x10000, math.sqrt(m.forwardVel) * 0x9000)
        set_character_anim_with_accel(m, CHAR_ANIM_FORWARD_SPINNING, accel)
        if _G.charSelect.gCSPlayers[m.playerIndex].modelId == E_MODEL_ARIO_BALL then
            smlua_anim_util_set_animation(m.marioObj, 'ario_ball_roll')
        else
            smlua_anim_util_set_animation(m.marioObj, 'ario_roll')
        end
    end

    set_mario_particle_flags(m, PARTICLE_SHALLOW_WATER_WAVE | PARTICLE_WAVE_TRAIL, 0)
    play_sound_with_freq_scale(SOUND_MOVING_TERRAIN_RIDING_SHELL + m.terrainSoundAddend,
        m.marioObj.header.gfx.cameraToObject, 1.0 + m.forwardVel / 64.0)

    m.actionTimer = m.actionTimer + 1
end
local function ario_water_roll_gravity(m)
    m.vel.y = 0
end

---@param m MarioState
local function ario_air_roll_loop(m)
    update_air_with_turn(m)

    if m.actionState == 0 then
        if m.controller.buttonPressed & Z_TRIG ~= 0 then
            m.actionState = 1
            play_sound(SOUND_GENERAL_SWISH_AIR_2, m.marioObj.header.gfx.cameraToObject)
        end
    elseif m.actionState == 1 then
        m.vel.y = -70.0
        set_mario_particle_flags(m, PARTICLE_DUST, 0)
        if m.controller.buttonPressed & Z_TRIG ~= 0 then
            return set_mario_action(m, ACT_GROUND_POUND, 0)
        end
    elseif m.actionState == 2 then -- delay bounce using state 2, helps with ground pound detection
        m.vel.y = 45.0
        m.actionState = 0
    end

    local step = perform_air_step(m, 0)
    if step == AIR_STEP_LANDED then
        if m.actionState == 0 then
            if m.forwardVel < 0 then
                -- fix negative vel bug
                mario_set_forward_vel(m, 0)
            end
            m.vel.y = 0
            return set_mario_action(m, ACT_ARIO_ROLL, 0)
        elseif m.actionState == 1 then
            m.actionState = 2

            play_mario_heavy_landing_sound(m, SOUND_ACTION_TERRAIN_HEAVY_LANDING)
            set_mario_particle_flags(m, PARTICLE_MIST_CIRCLE | PARTICLE_HORIZONTAL_STAR, 0)
        end

        check_fall_damage_or_get_stuck(m, ACT_HARD_FORWARD_GROUND_KB)
        return
    elseif step == AIR_STEP_HIT_WALL and not (m.wall == nil and gServerSettings.bouncyLevelBounds ~= BOUNCY_LEVEL_BOUNDS_OFF) then
        if m.forwardVel > 10.0 then
            mario_bonk_reflection(m, 0)
        else
            m.forwardVel = 0.0
        end
    end

    if m.actionTimer == 0 then
        if not gPlayerSyncTable[m.playerIndex].capThrown and m.flags & MARIO_NORMAL_CAP ~= 0 then
            if ario_enter_model_swap(m, 2) then
                spawn_mist_particles_variable(4, 1, 32.0)
                play_sound(SOUND_OBJ_BOO_BOUNCE_TOP, m.marioObj.header.gfx.cameraToObject)
            end
        end
    else
        local accel = max(0x10000, math.sqrt(m.forwardVel) * 0x8000)
        set_character_anim_with_accel(m, CHAR_ANIM_FORWARD_SPINNING, accel)
        if _G.charSelect.gCSPlayers[m.playerIndex].modelId == E_MODEL_ARIO_BALL then
            smlua_anim_util_set_animation(m.marioObj, 'ario_ball_roll')
        else
            smlua_anim_util_set_animation(m.marioObj, 'ario_roll')
        end
    end

    m.actionTimer = m.actionTimer + 1
end
local function ario_air_roll_gravity(m)
    if m.vel.y > 0 and m.controller.buttonDown & Z_TRIG == 0 then
        m.vel.y = maxf(0.0, m.vel.y - 4.0)
    end
    m.vel.y = maxf(-75.0, m.vel.y - 3.0)

    if m.pos.y < m.waterLevel - 100 then
        m.vel.y = m.vel.y * 0.75
    end
end

hook_mario_action(ACT_ARIO_ROLL, ario_roll_loop)
hook_mario_action(ACT_ARIO_WATER_ROLL, { every_frame = ario_water_roll_loop, gravity = ario_water_roll_gravity })
hook_mario_action(ACT_ARIO_AIR_ROLL, { every_frame = ario_air_roll_loop, gravity = ario_air_roll_gravity },
    INT_GROUND_POUND)


---@param m MarioState
local function ario_cap_hide_loop(m)
    if m.actionTimer == 0 then
        set_mario_particle_flags(m, PARTICLE_MIST_CIRCLE, 0)
        ario_enter_model_swap(m, 1)
        gStateExtras[m.playerIndex].capHideHits = 0
    end
    m.hurtCounter = 0

    m.forwardVel = approach_f32_symmetric(m.forwardVel, m.intendedMag / 4.0, 1.0)
    if m.input & INPUT_NONZERO_ANALOG ~= 0 then
        m.faceAngle.y = approach_s16_symmetric(m.faceAngle.y, m.intendedYaw, 0xA00)
    end

    if m.controller.buttonDown & (X_BUTTON | Z_TRIG) ~= (X_BUTTON | Z_TRIG) then
        play_sound(SOUND_OBJ2_MONTY_MOLE_APPEAR, m.marioObj.header.gfx.cameraToObject)
        return set_mario_action(m, ACT_STOP_CROUCHING, 0)
    end
    if m.controller.buttonPressed & A_BUTTON ~= 0 then
        return set_mario_action(m, ACT_BACKFLIP, 0)
    end
    if m.actionState > 3 then
        return set_mario_action(m, ACT_BACKWARD_GROUND_KB, 0)
    end

    mario_set_forward_vel(m, m.forwardVel)
    local step = perform_ground_step(m)

    local t = math.sin(m.actionTimer * 0.2) * m.intendedMag
    m.marioObj.header.gfx.angle.z = t * 0x40

    if step == GROUND_STEP_LEFT_GROUND then
        return set_mario_action(m, ACT_FREEFALL, 0)
    end

    -- ensure that model is reset
    if check_common_object_cancels(m) ~= 0 then
        set_mario_particle_flags(m, PARTICLE_MIST_CIRCLE, 0)
        ario_exit_model_swap(m)
    end

    m.actionTimer = m.actionTimer + 1
end

---@param m MarioState
local function ario_cap_hide_water_loop(m)
    if m.actionTimer == 0 then
        set_mario_particle_flags(m, PARTICLE_MIST_CIRCLE, 0)
        ario_enter_model_swap(m, 1)
        gStateExtras[m.playerIndex].capHideHits = 0
    end
    m.hurtCounter = 0

    m.forwardVel = approach_f32_symmetric(m.forwardVel, m.intendedMag / 4.0, 0.1)
    if m.input & INPUT_NONZERO_ANALOG ~= 0 then
        m.faceAngle.y = approach_s16_symmetric(m.faceAngle.y, m.intendedYaw, 0x200)
    end

    if m.controller.buttonDown & X_BUTTON == 0 or m.pos.y > m.waterLevel - 100 then
        play_sound(SOUND_OBJ2_MONTY_MOLE_APPEAR, m.marioObj.header.gfx.cameraToObject)
        return set_mario_action(m, ACT_WATER_IDLE, 0)
    end
    if m.actionState > 3 then
        return set_mario_action(m, ACT_BACKWARD_WATER_KB, 0)
    end

    mario_set_forward_vel(m, m.forwardVel)
    m.vel.y = -1.0
    if m.controller.buttonDown & A_BUTTON ~= 0 then
        m.vel.y = 3.0
    elseif m.controller.buttonDown & Z_TRIG ~= 0 then
        m.vel.y = -3.0
    end

    local step = perform_air_step(m, 0)

    local t = math.sin(m.actionTimer * 0x20)
    m.marioObj.header.gfx.pos.y = m.pos.y + t
    m.marioObj.header.gfx.angle.x = m.forwardVel * 0x200

    m.actionTimer = m.actionTimer + 1
end

hook_mario_action(ACT_ARIO_CAP_HIDE, ario_cap_hide_loop)
hook_mario_action(ACT_ARIO_CAP_HIDE_WATER, ario_cap_hide_water_loop)

local colObjLists = { OBJ_LIST_GENACTOR, OBJ_LIST_PUSHABLE, OBJ_LIST_SURFACE, OBJ_LIST_DESTRUCTIVE }

---@param o Object
local function inhale_delete(o)
    obj_mark_for_deletion(o)
end
---@param o Object
local function inhale_hide(o)
    o.header.gfx.node.flags = o.header.gfx.node.flags | GRAPH_RENDER_INVISIBLE
    o.oIntangibleTimer = -1
end
---@param o Object
local function inhale_bobomb(o)
    o.oAction = BOBOMB_ACT_DEATH_PLANE_DEATH
end
---@param o Object
local function inhale_piranhaplant(o)
    o.oPiranhaPlantScale = 0.0
    obj_scale(o, 0.0)
    o.oAction = PIRANHA_PLANT_ACT_WAIT_TO_RESPAWN
end
---@param o Object
local function inhale_amp(o)
    o.oPosX = o.oHomeX
    o.oPosY = o.oHomeY
    o.oPosZ = o.oHomeZ
    o.header.gfx.node.flags = GRAPH_RENDER_INVISIBLE
    o.oAction = HOMING_AMP_ACT_INACTIVE
    o.oAnimState = 0
    o.oForwardVel = 0
    o.oHomingAmpAvgY = o.oHomeY
end
---@param o Object
local function inhale_boo(o)
    if o.oBooParentBigBoo then
        o.oBooParentBigBoo.oBigBooNumMinionBoosKilled = o.oBooParentBigBoo.oBigBooNumMinionBoosKilled + 1
    end
    if o.oBehParams2ndByte ~= 0 then
        obj_mark_for_deletion(o)
    else
        -- ghost hunt dialogue
        o.oAction = 4
        local node = o.header.gfx.node
        node.flags = (node.flags & ~GRAPH_RENDER_ACTIVE) | GRAPH_RENDER_INVISIBLE
        o.oIntangibleTimer = -1
        o.oBooDeathStatus = BOO_DEATH_STATUS_DYING
    end
end
---@param o Object
local function inhale_pokey(o)
    local parent = o.parentObj
    parent.oPokeyNumAliveBodyParts = parent.oPokeyNumAliveBodyParts - 1
    if o.oBehParams2ndByte == 0 then
        parent.oPokeyHeadWasKilled = 1
    end
    parent.oPokeyAliveBodyPartFlags = parent.oPokeyAliveBodyPartFlags & ~(1 << o.oBehParams2ndByte)
    o.header.gfx.node.flags = o.header.gfx.node.flags | GRAPH_RENDER_INVISIBLE
end
---@param o Object
local function inhale_montymole(o)
    o.header.gfx.node.flags = o.header.gfx.node.flags | GRAPH_RENDER_INVISIBLE
    o.oIntangibleTimer = -1
    if not o.oMontyMoleCurrentHole then return end
    o.oMontyMoleCurrentHole.oMontyMoleHoleCooldown = 30
    o.oAction = MONTY_MOLE_ACT_SELECT_HOLE
    o.oVelY = 0.0
end
---@param o Object
local function inhale_bully(o)
    if o.oBullySubtype == BULLY_STYPE_MINION then
        o.parentObj.oBullyKBTimerAndMinionKOCounter = o.parentObj.oBullyKBTimerAndMinionKOCounter + 1
    end
    o.activeFlags = ACTIVE_FLAG_DEACTIVATED
end

local inhaleableBhvs = {}
hook_event(HOOK_ON_MODS_LOADED, function()
    inhaleableBhvs = {
        [id_bhvGoomba] = { fun = inhale_delete, mdl = E_MODEL_GOOMBA, anim = 'goomba_seg8_anims_0801DA4C' },
        [id_bhvBobomb] = { fun = inhale_bobomb, mdl = E_MODEL_BLACK_BOBOMB, anim = 'bobomb_seg8_anims_0802396C' },
        --[id_bhvPiranhaPlant] = { fun = inhale_piranhaplant, mdl = E_MODEL_PIRANHA_PLANT },
        [id_bhvKoopa] = { fun = inhale_delete, mdl = E_MODEL_KOOPA_WITHOUT_SHELL, anim = 'koopa_seg6_anims_06011364' },
        [id_bhvKoopaShell] = { fun = inhale_delete, mdl = E_MODEL_KOOPA_SHELL },
        [id_bhvSpindrift] = { fun = inhale_delete, mdl = E_MODEL_SPINDRIFT, anim = 'spindrift_seg5_anims_05002D68'  },
        [id_bhvMrBlizzardSnowball] = { fun = inhale_delete, bb = true },
        [id_bhvHomingAmp] = { fun = inhale_amp, mdl = E_MODEL_AMP, anim = 'amp_seg8_anims_08004034'  },
        [id_bhvCirclingAmp] = { fun = inhale_delete, mdl = E_MODEL_AMP, anim = 'amp_seg8_anims_08004034'  },
        [id_bhvEnemyLakitu] = { fun = inhale_delete, mdl = E_MODEL_ENEMY_LAKITU, anim = 'lakitu_enemy_seg5_anims_050144D4'  },
        [id_bhvSpiny] = { fun = inhale_delete, mdl = E_MODEL_SPINY_BALL },
        [id_bhvGhostHuntBoo] = { fun = inhale_boo, mdl = E_MODEL_BOO },
        [id_bhvMerryGoRoundBoo] = { fun = inhale_boo, mdl = E_MODEL_BOO },
        [id_bhvBoo] = { fun = inhale_delete, mdl = E_MODEL_BOO },
        [id_bhvFlyingBookend] = { fun = inhale_delete, mdl = E_MODEL_BOOKEND, anim = 'bookend_seg5_anims_05002540'  },
        [id_bhvHauntedChair] = { fun = inhale_delete, mdl = E_MODEL_HAUNTED_CHAIR },
        [id_bhvMoneybag] = { fun = inhale_delete, mdl = E_MODEL_MONEYBAG, anim = 'moneybag_seg6_anims_06005E5C'  },
        [id_bhvPokeyBodyPart] = { fun = inhale_pokey, mdl = E_MODEL_POKEY_BODY_PART, bb = true },
        [id_bhvFlyGuy] = { fun = inhale_delete, mdl = E_MODEL_FLYGUY, anim = 'flyguy_seg8_anims_08011A64'  },
        [id_bhvSwoop] = { fun = inhale_delete, mdl = E_MODEL_SWOOP, anim = 'swoop_seg6_anims_060070D0'  },
        [id_bhvMontyMole] = { fun = inhale_montymole, mdl = E_MODEL_MONTY_MOLE, anim = 'monty_mole_seg5_anims_05007248'  },
        [id_bhvMontyMoleRock] = { fun = inhale_delete, mdl = E_MODEL_PEBBLE, bb = true },
        [id_bhvMrIParticle] = { fun = inhale_delete, bb = true },
        [id_bhvSkeeter] = { fun = inhale_delete, mdl = E_MODEL_SKEETER, anim = 'skeeter_seg6_anims_06007DE0'  },
        [id_bhvScuttlebug] = { fun = inhale_delete, mdl = E_MODEL_SCUTTLEBUG, anim = 'scuttlebug_seg6_anims_06015064'  },
        [id_bhvSnufit] = { fun = inhale_delete, mdl = E_MODEL_SNUFIT },
        [id_bhvSnufitBalls] = { fun = inhale_delete, mdl = E_MODEL_METALLIC_BALL, bb = true },
        [id_bhvSmallBully] = { fun = inhale_bully, mdl = E_MODEL_BULLY, anim = 'bully_seg5_anims_0500470C'  },
        [id_bhvSmallChillBully] = { fun = inhale_bully, mdl = E_MODEL_CHILL_BULLY, anim = 'bully_seg5_anims_0500470C'  },
        [id_bhvFlameBouncing] = { fun = inhale_delete, mdl = E_MODEL_RED_FLAME, bb = true },
        [id_bhvFlameMovingForwardGrowing] = { fun = inhale_delete, mdl = E_MODEL_RED_FLAME, bb = true },
        [id_bhvBlueBowserFlame] = { fun = inhale_delete, mdl = E_MODEL_BLUE_FLAME, bb = true },
        [id_bhvSmallPiranhaFlame] = { fun = inhale_delete, mdl = E_MODEL_RED_FLAME, bb = true },
        [id_bhvBouncingFireballFlame] = { fun = inhale_delete, mdl = E_MODEL_RED_FLAME, bb = true },
        --[id_bhvWaterBomb] = { fun = inhale_delete, mdl = E_MODEL_WATER_BOMB },
        [id_bhvBombBit] = { fun = inhale_delete, mdl = E_MODEL_BOMB_BIT },
        [id_bhvCapInhaledObj] = { fun = inhale_delete },
    }
end)

local inhaleHitbox = { r = 150, h = 200 }

---@param o Object
---@param hitboxPos Vec3f
local function is_within_hitbox(o, hitboxPos, hitboxRadius, hitboxHeight)
    return (math.sqrt((hitboxPos.x - o.oPosX) ^ 2 + (hitboxPos.z - o.oPosZ) ^ 2) <= hitboxRadius + o.hitboxRadius)
        and (o.oPosY - o.hitboxDownOffset <= hitboxPos.y + hitboxHeight)
        and (o.oPosY + o.hitboxHeight - o.hitboxDownOffset >= hitboxPos.y)
end

---@param m MarioState
local function find_inhaled_obj(m)
    if m.playerIndex ~= 0 then return end

    local hitboxPos = {
        x = m.pos.x + sins(m.faceAngle.y) * 300,
        y = m.pos.y - 20,
        z = m.pos.z +
            coss(m.faceAngle.y) * 300
    }

    -- players
    for i = 0, MAX_PLAYERS - 1, 1 do
        if gNetworkPlayers[i].connected then
            local m2 = gMarioStates[i]
            if m2 and m2.playerIndex ~= 0
                and m2.action & (ACT_FLAG_INVULNERABLE | ACT_FLAG_INTANGIBLE) == 0 and m.invincTimer == 0
                and is_within_hitbox(m2.marioObj, hitboxPos, inhaleHitbox.r, inhaleHitbox.h) then
                network_send_to(m2.playerIndex, true, { id = PACKET_INHALED, i = m.marioObj.globalPlayerIndex })
                return true
            end
        end
    end

    -- other objects
    for i, list in ipairs(colObjLists) do
        local o2 = obj_get_first(list)
        while o2 do
            local bhv = inhaleableBhvs[get_id_from_behavior(o2.behavior)]
            if bhv and o2.oIntangibleTimer == 0 then
                if is_within_hitbox(o2, hitboxPos, inhaleHitbox.r, inhaleHitbox.h) then
                    local model = bhv.mdl or obj_get_model_id_extended(o2)
                    local scale = o2.header.gfx.scale.x
                    spawn_sync_object(id_bhvCapInhaledObj, model, o2.oPosX, o2.oPosY, o2.oPosZ, function(o)
                        o.oInhalerIndex = m.marioObj.globalPlayerIndex
                        obj_scale(o, scale)
                        o.oInhaledObjScale = scale
                        o.oInhaledObjModel = model
                        if bhv.bb then obj_set_billboard(o) end
                        if bhv.anim then o.oAnimations = gObjectAnimations[bhv.anim] end

                        if not o2 then return end
                        o.oAnimations = o2.oAnimations
                        o.oOpacity = o2.oOpacity
                        o.oMoveAngleYaw = obj_angle_to_point(o, m.pos.x, m.pos.z)
                    end)
                    bhv.fun(o2)
                    network_send_object(o2, true)
                    return true
                end
            end
            o2 = obj_get_next(o2)
        end
    end
end

local function spawn_inhale_particle(m)
    local particle = spawn_non_sync_object(id_bhvWind, E_MODEL_MIST, m.pos.x + sins(m.faceAngle.y) * 500,
        m.pos.y + 80, m.pos.z + coss(m.faceAngle.y) * 500, nil)
    particle.oTimer = 3
    obj_translate_xyz_random(particle, 200.0)
    particle.oMoveAngleYaw = obj_angle_to_object(particle, m.marioObj)
    particle.oForwardVel = random_float() * 30.0 + 40.0
    particle.oVelY = (m.pos.y + 80 - particle.oPosY) / 4
    obj_set_billboard(particle)
    particle.oOpacity = 100
end

local function spawn_held_cap_obj(m)
    if m.playerIndex == 0 then
        local nearestCap = obj_get_nearest_object_with_behavior_id(m.marioObj, id_bhvArioHeldCap)
        if nearestCap and nearestCap.globalPlayerIndex == m.marioObj.globalPlayerIndex then return end

        local alt = _G.charSelect.character_get_current_costume(0)
        spawn_sync_object(id_bhvArioHeldCap, alt == 1 and CAPMODELS_ARIO_CLOUD.normal or CAPMODELS_ARIO.normal, m.pos.x, m.pos.y, m.pos.z, function(o)
            o.globalPlayerIndex = m.marioObj.globalPlayerIndex
            o.header.gfx.node.flags = o.header.gfx.node.flags | GRAPH_RENDER_INVISIBLE
        end)
    end
end

---@param m MarioState
local function act_ario_cap_inhale(m)
    if m.actionArg == 0 then
        set_character_animation(m, CHAR_ANIM_PICK_UP_LIGHT_OBJ)
        if m.actionTimer == 0 then
            spawn_held_cap_obj(m)

            set_anim_to_frame(m, 5)
            m.actionTimer = m.actionTimer + 1
        end

        mario_set_forward_vel(m, approach_f32_symmetric(m.forwardVel, 0.0, 4.0))

        if is_anim_at_end(m) ~= 0 then
            m.actionArg = 1
            m.actionTimer = 0
        end
    elseif m.actionArg == 1 then
        play_sound_with_freq_scale(SOUND_AIR_HOWLING_WIND, m.marioObj.header.gfx.cameraToObject, 1.8)
        spawn_inhale_particle(m)

        m.intendedMag = m.intendedMag * 0.4

        update_walking_speed(m)

        if m.forwardVel <= 2.0 and m.input & INPUT_NONZERO_ANALOG == 0 then
            mario_set_forward_vel(m, 0.0)
            set_character_animation(m, CHAR_ANIM_IDLE_WITH_LIGHT_OBJ)
        else
            anim_and_audio_for_hold_walk(m)
        end

        if m.controller.buttonDown & L_TRIG == 0 then
            m.actionArg = 2
            m.actionTimer = 55
        end
        if m.controller.buttonPressed & A_BUTTON ~= 0 then
            m.vel.y = 40.0
            return set_mario_action(m, ACT_ARIO_CAP_INHALE_AIR, 1)
        end

        if find_inhaled_obj(m) then
            m.actionArg = 2
            m.actionTimer = 0
        end
    else
        set_character_animation(m, CHAR_ANIM_IDLE_WITH_LIGHT_OBJ)
        spawn_inhale_particle(m)

        m.actionTimer = m.actionTimer + 1

        mario_set_forward_vel(m, 0.0)

        -- failsafe for if the inhaled obj unloads or something
        if m.actionTimer >= 60 then
            m.marioBodyState.handState = MARIO_HAND_FISTS
            return set_mario_action(m, ACT_IDLE, 0)
        end
    end

    local step = perform_ground_step(m)

    if step == GROUND_STEP_LEFT_GROUND then
        return drop_and_set_mario_action(m, ACT_FREEFALL, 0)
    elseif step == GROUND_STEP_HIT_WALL then
        if m.forwardVel > 16.0 then
            mario_set_forward_vel(m, 16.0)
        end
    end
end

---@param m MarioState
local function act_ario_cap_inhale_air(m)
    set_character_animation(m, CHAR_ANIM_JUMP_WITH_LIGHT_OBJ)

    if m.actionArg == 0 then
        if m.actionTimer == 0 then
            spawn_held_cap_obj(m)
        end

        if m.actionTimer > 5 then
            m.actionArg = 1
            m.actionTimer = 0
        end

        m.actionTimer = m.actionTimer + 1
    elseif m.actionArg == 1 then
        play_sound_with_freq_scale(SOUND_AIR_HOWLING_WIND, m.marioObj.header.gfx.cameraToObject, 1.8)
        spawn_inhale_particle(m)

        m.intendedMag = m.intendedMag * 0.5

        if m.controller.buttonDown & L_TRIG == 0 then
            m.actionArg = 2
            m.actionTimer = 55
        end

        if find_inhaled_obj(m) then
            m.actionArg = 2
            m.actionTimer = 0
        end
    else
        spawn_inhale_particle(m)

        m.actionTimer = m.actionTimer + 1

        mario_set_forward_vel(m, approach_f32_symmetric(m.forwardVel, 0.0, 4.0))

        -- failsafe for if the inhaled obj unloads or something
        if m.actionTimer >= 60 then
            m.marioBodyState.handState = MARIO_HAND_FISTS
            return set_mario_action(m, ACT_FREEFALL, 0)
        end
    end

    if m.vel.y > 16.0 and m.controller.buttonDown & A_BUTTON == 0 then
        m.vel.y = 16.0
    end

    update_air_without_turn(m)
    local step = perform_air_step(m, 0)

    if step == AIR_STEP_LANDED then
        if check_fall_damage_or_get_stuck(m, ACT_HARD_FORWARD_GROUND_KB) == 0 then
            local timer = m.actionTimer
            set_mario_action(m, ACT_ARIO_CAP_INHALE, m.actionArg)
            m.actionTimer = timer
        end
        return
    elseif step == AIR_STEP_HIT_WALL then
        if m.forwardVel > 16.0 then
            mario_set_forward_vel(m, 16.0)
        end
    end
end

hook_mario_action(ACT_ARIO_CAP_INHALE, act_ario_cap_inhale)
hook_mario_action(ACT_ARIO_CAP_INHALE_AIR, act_ario_cap_inhale_air)

---@param m MarioState
local function act_inhaled(m)
    local m2 = gMarioStates[network_local_index_from_global(m.actionArg)]
    local e = gStateExtras[m.playerIndex]
    set_character_animation(m, CHAR_ANIM_FLY_FROM_CANNON)

    if m.actionState == 0 then
        if m.actionTimer == 0 then
            play_character_sound(m, CHAR_SOUND_WHOA)
        end

        m.faceAngle.y = obj_angle_to_object(m.marioObj, m2.marioObj)
        mario_set_forward_vel(m, approach_f32_symmetric(m.forwardVel, 40.0, 5.0))
        m.vel.y = 0.0
        m.pos.y = approach_f32_asymptotic(m.pos.y, m2.pos.y + 80, 1.5)

        perform_air_step(m, 0)

        if m2.action ~= ACT_ARIO_CAP_INHALE and m.action ~= ACT_ARIO_CAP_INHALE_AIR then
            return drop_and_set_mario_action(m, ACT_FORWARD_AIR_KB, 0)
        end

        if lateral_dist_between_objects(m.marioObj, m2.marioObj) < 100 or m.actionTimer > 50 then
            if m.playerIndex == 0 then
                network_send_to(m2.playerIndex, true, { id = PACKET_INHALED_END, i = network_global_index_from_local(0) })
            end
            m.actionState = 1
        end
        m.actionTimer = m.actionTimer + 1
    else
        local node = m.marioObj.header.gfx.node
        node.flags = node.flags | GRAPH_RENDER_INVISIBLE
        vec3f_copy(m.pos, m2.pos)

        e.inhaleEscapeCounter = max(0, e.inhaleEscapeCounter - 1)
        if m.input & INPUT_A_PRESSED ~= 0 then
            e.inhaleEscapeCounter = e.inhaleEscapeCounter + 10
            if e.inhaleEscapeCounter > 40 then
                m.faceAngle.y = m2.faceAngle.y
                m.pos.y = m.pos.y + 50
                m.forwardVel = 10.0
                e.inhaleEscapeCounter = 0

                play_sound(SOUND_ACTION_CLIMB_UP_POLE, m.marioObj.header.gfx.cameraToObject)

                set_mario_action(m, ACT_FORWARD_AIR_KB, 0)

                if m.playerIndex == 0 then
                    network_send_to(m2.playerIndex, true, { id = PACKET_ESCAPED_INHALE, i = m.marioObj.globalPlayerIndex })
                end
            end
        end
    end
end

hook_mario_action(ACT_INHALED, act_inhaled)
